/*
 * aconvert.cpp:
 *
 * Read a raw image file and write an .aff file.
 */

/* Be sure we have 64-bit support */

#include "afflib.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <signal.h>
#include <term.h>

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Simson L. Garfinkel, 
 *      Basis Technology, and its contributors.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */





char *progname = "aconvert";

int	data_segsize = 16*1024*1024;	// default seg size --- 16MB
int	opt_compress	= 1;
int	opt_fake	= 0;
int	fake_segments	= 0;			// how many fake segments we need
long long bytes_to_convert = 0;
int	opt_batch	= 1;
char	*opt_ext	= "aff";
int	opt_delete_source = 0;
int	opt_quiet	= 0;
int	opt_nozprobe	= 0;
int	opt_revert	= 0;

void usage()
{
    printf("usage: %s [options] raw1 raw2 raw3...  --- batch convert each raw file\n",progname);
    printf("      -s nnnn  -- set the data_segsize (default %d)\n",data_segsize);
    printf("      -x       -- don't compress AFF file (mostly for testing).\n");
    printf("      -g nnn   -- fake input of nnn gigabytes (for testing)\n");
    printf("      -e ext   -- use 'ext' for the converted files (default %s)\n",opt_ext);
    printf("      -w file  -- convert to 'file' (can only convert one at a time)\n");
    printf("      -z       -- zap; delete each fsource ile after it has been converted.\n");
    printf("      -q       -- Quiet mode. Don't ask questions, don't print status.\n");
    printf("      -Z       -- Do not automatically probe for gzip/bzip2 compression.\n");
    printf("      -r       -- REVERT the files; if they are .aff files, convert back.\n");
    printf("                  Sets '-e .ext' but you can change it.\n");
    printf("\n");
    exit(0);
}


AFFILE *af = 0;				// file being converted
void cleanup()
{
    fprintf(stderr,"Erasing %s\n",af->fname);
    unlink(af->fname);
    exit(1);
}

void sig_intr(int val)
{
    fprintf(stderr,"\n\n^C interrupt. ");
    if(af) cleanup();
    exit(1);
}



/* probe_gzip():
 * Is this a gzip file?
 * Right now it just looks at the file extension.
 */

int probe_gzip(const char *infile)
{
    int len = strlen(infile);

    if(len>3 && strcmp(infile+len-3,".gz")==0){
	return 1;
    }
    return 0;
}

int probe_bzip2(const char *infile)
{
    int len = strlen(infile);

    if(len>4 && strcmp(infile+len-4,".bz2")==0){
	return 1;
    }
    return 0;
}

/* yesno():
 * As a yes/no question. Return 1 if yes, 0 if no.
 */

int yesno(char *prompt)
{
    char buf[256];
    do {
	printf("%s [y/n]: ",prompt);
	memset(buf,0,sizeof(buf));
	fgets(buf,sizeof(buf)-1,stdin);
	if(buf[0]=='y' || buf[0]=='Y') return 1;
    } while(buf[0]!='n' && buf[0]!='N');
    return 0;
}

int revert(char *infile,char *outfile)
{
    /* If infile is not an AFF file, return */
    if(af_identify(infile)!=1){
	fprintf(stderr,"%s is not an AFF file.\n",infile);
	return -1;
    }
    /* if outfile is NULL, figure it out from infile */
    if(outfile==0){
	outfile = (char *)alloca(strlen(infile)+16);
	strcpy(outfile,infile);
	char *cc = rindex(outfile,'.'); // to strip off extension
	if(cc){
	    strcpy(cc+1,opt_ext);
	}
	else {
	    strcat(outfile,".");
	    strcat(outfile,opt_ext);
	}
    }
    struct stat so;
    if(stat(outfile,&so)==0){
	fprintf(stderr,"%s: file exists. Delete it before reverting.\n",outfile);
	return -1;
    }
    printf("converting %s to %s...\n",infile,outfile);
    AFFILE *af = af_open(infile,O_RDONLY,0);
    if(!af){
	warn("af_open(%s) ",infile);
	return -1;
    }
    if(af->image_datasegsize==0){
	fprintf(stderr,"image datasegsize not set.\n");
	af_close(af);
	return -1;
    }
    /* Now, rewind to the beginning of the AF file and step through each segment... */
    af_rewindi(af);

    /* Open the output file */
    int ofd = open(outfile,O_WRONLY|O_CREAT|O_TRUNC,0666);
    if(ofd<0) err(1,outfile);

    char segname[AF_MAX_NAME_LEN];
    size_t data_len = 0;
    unsigned char *data;
    unsigned long arg;
    while(af_get_next_seg(af,segname,sizeof(segname),&arg,&data,&data_len)==0){
	unsigned int segnum=0;
	printf("Segment: %-20s",segname);
	tputs(clr_eol,0,putchar);
	if(sscanf(segname,AF_SEG,&segnum)==1){
	    if(arg & AF_SEG_COMPRESSED){
		Bytef *ndata = (unsigned char *)malloc(af->image_datasegsize);

		uLongf uncompressed_bytes=af->image_datasegsize;

		printf("UNCOMPRESSING...");
		int r = uncompress(ndata,&uncompressed_bytes,(const Bytef *)data,data_len);
		switch(r){
		case Z_OK:
		    data_len = uncompressed_bytes;
		    free(data);		// don't need this old one anymore
		    data = ndata;
		    break;
		case Z_MEM_ERROR:
		    err(1,"Z_MEM_ERROR decompressing segment %d",segnum);
		case Z_DATA_ERROR:
		    err(2,"Z_DATA_ERROR decompressing segment %d",segnum);
		default:
		    err(3,"uncompress returned an invalid value in get_segment");
		}
		printf("\010\010\010\010\010\010\010\010\010\010\010\010\010\010\010\010");
		tputs(clr_eol,0,putchar);
	    }
	    off_t npos = af->image_datasegsize * segnum;
	    printf("writing at offset=%qd len=%d",(unsigned long long)npos,(int)data_len);
	    lseek(ofd,npos,SEEK_SET);
	    if(write(ofd,data,data_len)!=(int)data_len){
		err(1,"write(%s):",outfile);
	    }
	}
	free(data);
	printf("\n");
    }
    af_close(af);
    return 0;
}


int convert(const char *infile,char *outfile)
{
    /* if infile is an AFF file, don't convert it... */
    if(af_identify(infile)==1){
	fprintf(stderr,"%s is an AFF file; will not convert.\n",infile);
	return -1;
    }

    /* if outfile is NULL, figure it out from infile */
    if(outfile==0){
	outfile = (char *)alloca(strlen(infile)+16);
	strcpy(outfile,infile);
	char *cc = rindex(outfile,'.'); // to strip off extension
	if(cc){
	    strcpy(cc+1,opt_ext);
	}
	else {
	    strcat(outfile,".");
	    strcat(outfile,opt_ext);
	}
    }

    struct stat so;
    if(stat(outfile,&so)==0){
	fprintf(stderr,"%s: file exists. Delete it before converting.\n",outfile);
	return -1;
    }

    /* Figure out how many bytes to convert if we are not faking it*/
    struct stat si;
    if(stat(infile,&si)){
	err(2,"stat(%s) ",infile);
    }
    if(!opt_fake){
	bytes_to_convert = si.st_size;
    } else {
	bytes_to_convert = (unsigned long long)fake_segments * data_segsize;
    }


    FILE *fin = 0;
    int  fin_popen = 0;			// was this opened with popen?

    /* Check to see if it is a gzip file... */
    if(probe_gzip(infile) && yesno("infile looks like a gzip file; Uncompress it?")){
	/* Open with a subprocess. We will need to use zlib when we move to Windows. */
	char buf[256];
	sprintf(buf,"gzcat %s",infile);
	fin = popen(buf,"r");
	fin_popen = 1;
    }

    /* Check to see if it is a bzip2 file... */
    if(probe_bzip2(infile) && yesno("infile looks like a bzip2 file; Uncompress it?")){
	/* Open with a subprocess. We will need to use bzip2zlib when we move to Windows. */
	char buf[256];
	sprintf(buf,"bzcat %s",infile);
	fin = popen(buf,"r");
	fin_popen = 1;
    }

    if(fin==0){
	/* Looks like it's a regular file */
	printf("converting %s to %s...\n",infile,outfile);
	 
	fin = fopen(infile,"rb");
	if(!fin){
	    err(1,"fopen %s: ",infile);
	}
    }


    /* Create the AFF file */
    af = af_open(outfile,O_CREAT|O_RDWR|O_TRUNC,0666);
    if(!af){
	err(1,"af_open %s: ",outfile);
    }

    /* Write out the segment size */
    af_set_sizes(af,data_segsize,(unsigned long long)0);		// 

    /* Store empty hashes, which will be updated later. */
    unsigned char md5_buf[16];
    unsigned char sha1_buf[20];

    memset(md5_buf,0,sizeof(md5_buf));
    memset(sha1_buf,0,sizeof(sha1_buf));
    af_put_seg(af,AF_MD5,0,md5_buf,16);
    af_put_seg(af,AF_SHA1,0,sha1_buf,20);

    /* Now we are done with metadata; start the conversion... */

    char *data = (char *)malloc(data_segsize); // buffer to store the data we read
    char *cdata = (char *)malloc(data_segsize); // buffer to store the compressed data
    if(!data || !cdata){
	err(1,"malloc");
    }
    int segnum = 0;
    long long bytes_read=0;
    long long bytes_written=0;
    long csegs_written = 0;
    long segs_written = 0;

    /* Set up the MD5 machinery */
    MD5_CTX md5;
    MD5_Init(&md5);

    SHA_CTX sha;
    SHA1_Init(&sha);

    while(!feof(fin)){
	printf("\rConverting segment %d  ",segnum);
	printf("%qd/%qd bytes ",bytes_read,bytes_to_convert);
	printf("%6.2f%% ...",bytes_read*100.0/bytes_to_convert);
	fflush(stdout);

	size_t bytes = 0;

	/* Read the data from the file to convert */
	if(opt_fake==0){
	    bytes = fread(data,1,data_segsize,fin);
	}
	else {
	    bytes = data_segsize;
	    fake_segments--;
	    if(fake_segments==0) bytes = 0;
	    sprintf(data,"fake segment %d\n",segnum);
	}

	if(bytes<=0) break;

	MD5_Update(&md5,data,bytes);
	SHA1_Update(&sha,data,bytes);
	bytes_read += bytes;

	/* Compress the data */
	uLongf destLen = data_segsize;
	int cres = -1;			// compression result; default is that we failed

	if(opt_compress){		// try to compress
	    cres = compress((Bytef *)cdata,&destLen,(Bytef *)data,bytes);
	}

	int r = 0;
	char vbuf[256];
	sprintf(vbuf,AF_SEG,segnum);
	if(cres==0 && destLen < bytes){
	    /* Prepare to write out the compressed segment with compression*/
	    r = af_put_seg(af,vbuf,AF_SEG_COMPRESSED,cdata,destLen);
	    bytes_written += destLen;
	    csegs_written++;
	}
	else {
	    /* Prepare to write out the segment without compression */
	    r = af_put_seg(af,vbuf,0,data,bytes);
	    bytes_written += bytes;
	    segs_written++;
	}
	if(r){
	    fprintf(stderr,"Error writing output file.\n");
	    cleanup();
	}

	segnum++;
    }
    if(fin_popen){
	pclose(fin);
    }
    else {
	fclose(fin);
    }


    printf("\n\nDone!\n\n");

    /* Go back and update the image size */
    if(af_update_segq(af,AF_IMAGESIZE,bytes_read,0)){
	err(1,"Could not upate AF_IMAGESIZE");
    }

    /* Finish the hash calculations and write to the db */
    MD5_Final(md5_buf,&md5);
    char buf[256];
    if(af_update_seg(af,AF_MD5,0,md5_buf,16,0)){
	err(1,"Could not update AF_MD5");
    }
    printf("md5: %s\n",af_hexbuf(buf,sizeof(buf),md5_buf,16,1));

    SHA1_Final(sha1_buf,&sha);
    if(af_update_seg(af,AF_SHA1,0,sha1_buf,20,0)){
	err(1,"Could not update AF_SHA1");
    }

    printf("sha1: %s\n",af_hexbuf(buf,sizeof(buf),sha1_buf,20,1));
    af_close(af);
    af = 0;

    if(stat(outfile,&so)){
	err(1,"stat(%s) ",outfile);
    }

    printf("bytes read: %qd  bytes written: %qd\n",bytes_read,bytes_written);
    printf("Total segments: %d  (%ld compressed)\n",segnum,csegs_written);
    printf("Conversion finished.\n");
    printf("AFF file is %qd bytes. Net compression ratio with overhead: %4.1f%%\n",
	   (long long)so.st_size,((float)bytes_read-so.st_size)*100.0/bytes_read);
    printf("\n\n");

    /* Set the utime on the resulting file */
    struct timeval times[2];
    memset(&times,0,sizeof(times));
    times[0].tv_sec = si.st_mtime;
    times[1].tv_sec = si.st_mtime;
    utimes(outfile,times);
    if(opt_delete_source){
	printf("DELETE %s\n",infile);
	if(unlink(infile)){
	    perror(infile);
	}
    }
    return(0);
}


int main(int argc,char **argv)
{
    char *outfile = 0;
    int ch;

    setupterm((char *)0,1,(int *)0);
    signal(SIGINT,sig_intr);

    progname = argv[0];
    while ((ch = getopt(argc, argv, "s:xg:h?ew:zqr")) != -1) {
	switch (ch) {
	case 'g':
	    opt_fake = 1;
	    fake_segments = (1024*1024*1024 / data_segsize) * atoi(optarg);
	    printf("Will need %d fake segments...\n",fake_segments);
	    break;
	case 'e':
	    opt_ext = optarg;
	    break;
	case 'w':
	    outfile = optarg;
	    break;
	case 'z':
	    opt_delete_source ++;
	    break;
	case 'q':
	    opt_quiet++;
	    break;
	case 'r':
	    opt_revert++;
	    opt_ext = "raw";
	    break;
	case 's':
	    data_segsize = atoi(optarg);
	    break;
	case 'x':
	    opt_compress=0;
	    break;
	case 'Z':
	    opt_nozprobe++;
	    break;

	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }

    argc--;

    if(outfile){
	convert(*argv,outfile);
	return 0;			// only can convert a single file
    }

    while(*argv){
	if(opt_revert==0){
	    convert(*argv++,0);
	}
	else {
	    revert(*argv++,0);
	}
	argc--;
    }
    return 0;
}
