/* hexbuf(char *dest,void *bin,int bytes):
 * convert some binary data to hex, usually for printing.
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


/* Be sure we have 64-bit support */
#ifndef LARGEFILE_SOURCE
#define LARGEFILE_SOURCE 1
#endif

#ifndef _FILE_OFFSET_BITS
#define FILE_OFFSET_BITS 64
#endif


#include "afflib.h"
#include <stdio.h>
#include <err.h>
#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <openssl/rand.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>

/* Utility Functions */

#ifdef __APPLE__
#define POINTER_FMT "%p"
#endif

#ifdef linux
#define POINTER_FMT "%p"
#endif

#ifndef POINTER_FMT
#define POINTER_FMT "%x"
#endif


/* Internal Functions */
void	af_get_segsizes(AFFILE *af);	// sets up values if we can get them.
void	af_demand_segsizes(AFFILE *af);	// abort if we can't get them



/****************************************************************
 *** low-level functions for putting...
 ****************************************************************/

/* af_put_seg:
 * put the given named segment at the current position in the file.
 * Return 0 for success, -1 for failure (probably disk full?)
 * TODO: Remove the incomplete transaction when writing...
 */

int af_put_seg(AFFILE *af, const char *name,unsigned long arg,const void *data,unsigned int datalen)
{
    struct af_segment_head segh;
    struct af_segment_tail segt;

    if(af->debug){
	fprintf(stderr,"af_put_seg(" POINTER_FMT ",'%s',%lu,data=" POINTER_FMT ",datalen=%u)\n",
		af,name,arg,data,datalen);
    }

    assert(sizeof(segh)==16);
    assert(sizeof(segt)==8);

    /* If the last command was not a probe (so we know where we are), and
     * we are not at the end of the file, something is very wrong.
     */
    //    if(af->last_command != 1 && !feof(af->aseg)){
    //	fprintf(stderr,"af_put_segv: last_command=%d  feof=%d\n",af->last_command,feof(af->aseg));
    //exit(1);
    //}

    unsigned int namelen = strlen(name);

    strcpy(segh.magic,AF_SEGHEAD);
    segh.name_len = htonl(namelen);
    segh.data_len = htonl(datalen);
    segh.flag      = htonl(arg);

    strcpy(segt.magic,AF_SEGTAIL);
    segt.segment_len = htonl(sizeof(segh)+namelen + datalen + sizeof(segt));
    
#ifdef DEBUG   
    {
	aff_ull o = ftello(af->aseg);
	fprintf(stderr,"af_put_segv: putting segment %s (datalen=%d) offset=%qd\n",
		name,datalen,o);
    }
#endif

    if(fwrite(&segh,sizeof(segh),1,af->aseg)!=1) return -1;
    if(fwrite(name,1,namelen,af->aseg)!=namelen) return -1;
    if(fwrite(data,1,datalen,af->aseg)!=datalen) return -1;
    if(fwrite(&segt,sizeof(segt),1,af->aseg)!=1) return -1;
    fflush(af->aseg);			// make sure it is on the disk
    af->last_command = 2;
    return 0;
}



/* af_put_segq:
 * Put the named quad-byte valud at the current position in the file.
 */

int af_put_segq(AFFILE *af,const char *name,unsigned long long value)
{
    struct aff_quad  q;

    assert(sizeof(q)==8);

    q.low  = htonl(value & 0xffffffff);
    q.high = htonl(value >> 32);
    
    return af_put_seg(af,name,AF_SEG_QUADWORD,&q,8);
}

/****************************************************************
 *** Probe Functions 
 ****************************************************************/

int af_probe_next_seg(AFFILE *af,
		       char *segname,
		       size_t segname_len,
		       unsigned long *arg_, // optional arg
		       size_t *datasize_, // optional get datasize
		       size_t *segsize_, // optional get size of entire segment
		       int do_rewind) // optional rewind af->aseg, otherwise leave at data
{
    if(!af->aseg) err(1,"af_probe_segment only works with aff files");    

    struct af_segment_head segh = {0,0};

    aff_ull start = ftello(af->aseg);

    if(fread(&segh,sizeof(segh),1,af->aseg)!=1){
	return -1;
    }
    if(strcmp(segh.magic,AF_SEGHEAD)!=0){
	err(1,"afflib: segh is corrupt at %qd\n",start);
    }

    unsigned long name_len = ntohl(segh.name_len);
    unsigned long datasize = ntohl(segh.data_len);
    if(name_len>AF_MAX_NAME_LEN){
	err(1,"afflib: name_len=%lu (an outrageous value)",name_len);
    }

    if(name_len+1 > segname_len){
	fseeko(af->aseg,start,SEEK_SET); // rewind to start
	return -2;
    }

    if(fread(segname,1,name_len,af->aseg)!=name_len){
	fseeko(af->aseg,start,SEEK_SET); // rewind
	return -1;
    }
    segname[name_len] = 0;

    if(do_rewind) fseeko(af->aseg,start,SEEK_SET); // rewind

    unsigned long segsize = sizeof(struct af_segment_head)
	+ sizeof(struct af_segment_tail)
	+ name_len + datasize;

    if(arg_)      *arg_      = ntohl(segh.flag);
    if(datasize_) *datasize_ = datasize;
    if(segsize_)  *segsize_ = segsize;

#ifdef DEBUG
    fprintf(stderr,"af_probe_next_seg(segname=%s datasize=%d segsize=%d) do_rewind=%d\n",
	    segname,datasize,segsize,do_rewind);
#endif
    af->last_command = 1;
    return 0;
}



/* find the given segment and return 0 if found.
 * Leave the file pointer positioned at the start of the segment.
 * Return -1 if segment is not found, and leave pointer at the end
 */
int	af_find_seg(AFFILE *af,const char *segname,
		    unsigned long *arg,
		    size_t *datasize,
		    size_t *segsize)
{

    af_rewindi(af);

    char   next_segment_name[AF_MAX_NAME_LEN];
    size_t next_segsize = 0;
    size_t next_datasize = 0;
    unsigned long next_arg;
    
    while(af_probe_next_seg(af,next_segment_name,sizeof(next_segment_name),
			    &next_arg,&next_datasize,&next_segsize,1)==0){
	if(strcmp(next_segment_name,segname)==0){ // found the segment!
	    if(datasize) *datasize = next_datasize;
	    if(segsize)  *segsize  = next_segsize;
	    if(arg)      *arg      = next_arg;
	    return 0;			// return the info
	}
	fseeko(af->aseg,next_segsize,SEEK_CUR);	// skip the segment
    }
    return -1;				// couldn't find segment
}
		    

/* af_write_ignore:
 * write an AF_IGNORE structure of a given size.
 */
   

int af_ignore_overhead()
{
    return sizeof(struct af_segment_head)+sizeof(struct af_segment_tail);
}

int af_write_ignore(AFFILE *af,unsigned long bytes)
{

    unsigned char *invalidate_data = (unsigned char *)calloc(bytes,1);

    af_put_seg(af,AF_IGNORE,0,invalidate_data,bytes); // overwrite with NULLs
    free(invalidate_data);
    return(0);
}

/* Delete the first occurance of the named segment */
int	af_del_seg(AFFILE *af,const char *segname)
{
    size_t segsize;
    if(af_find_seg(af,segname,0,0,&segsize)!=0){
	return -1;
    }
    /* Now wipe it out */
    size_t ignore_size = segsize-strlen(segname);
    af_write_ignore(af,ignore_size);
    return 0;
}



/****************************************************************
 *** GET FUNCTIONS
 ****************************************************************/

/****************************************************************
 *** Probe the next segment:
 *** Return its name and argument, but don't advance the pointer....
 *** Returns 0 on success, -1 on end of file or other error.
 ****************************************************************/
/****************************************************************
 *** Get the next segment.
 *** data must be freed.
 *** if no data is specified, then just seek past it...
 *** Returns 0 on success, -1 on end of file or other error.
 ****************************************************************/
int af_get_next_seg(AFFILE *af,char *segname,size_t segname_len,unsigned long *arg,
			unsigned char **data,size_t *datalen_)
{
    if(!af->aseg) err(1,"af_get_next_segv only works with aff files");

    size_t data_len;
    if(af_probe_next_seg(af,segname,segname_len,arg,&data_len,0,0)){
	return -1;			// must be end of file or something
    }
    if(data){
	/* Allocate the data and read... */
	*data    = (unsigned char *)malloc(data_len);
	if(fread(*data,1,data_len,af->aseg)!=data_len){
	    fprintf(stderr,"af_get_next_segv: EOF on reading segment? File is corrupt.\n");
	    exit(1);
	}
    } else {
	fseeko(af->aseg,data_len,SEEK_CUR); // skip past the data
    }
    if(datalen_) *datalen_ = data_len;

    /* Now read the tail */
    struct af_segment_tail segt;
    if(fread(&segt,sizeof(segt),1,af->aseg)!=1){
	fprintf(stderr,"af_get_next_segv: end of file reading segment tail...\n");
	exit(1);
    }
    /* Validate tail */
    unsigned long stl = ntohl(segt.segment_len);
    unsigned long calculated_segment_len =
	sizeof(struct af_segment_head)
	+ strlen(segname)
	+ data_len + sizeof(struct af_segment_tail);

    if(strcmp(segt.magic,AF_SEGTAIL)!=0){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt. No tail!\n");
	exit(1);
    }
    if(stl != calculated_segment_len){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt reading tail (%lu!=%lu)!\n",
		stl,calculated_segment_len);
	exit(1);
    }
    return 0;
}


/* af_get_segment:
 * Get the named segment.
 * If it is the next segment, just get it.
 * Otherwise go to the beginning and scan for it.
 * Eventually we hope to cache the names and starting locations of each segment...
 */

int af_get_seg(AFFILE *af,const char *name,unsigned long *arg,
		unsigned char **data,size_t *datalen)
{
    char next[AF_MAX_NAME_LEN];
    int first = 1;
    size_t segsize = 0;
    
    do {

	if(af_probe_next_seg(af,next,sizeof(next),0,0,&segsize,1)==0){	
	    if(strcmp(next,name)==0){	// found the segment!
		int ret = af_get_next_seg(af,next,sizeof(next),arg,data,datalen);
		assert(strcmp(next,name)==0);	// hopefully it's still the same!
		return ret;
	    }
	}
	else {
	    /* Must be at EOF. If this is not first time through,
	     * then segment does not exist...
	     */
	    if(!first) break;
	}
	if(first){
	    /* If first time we checked, rewind to the first segment */
	    af_rewindi(af);
	    first = 0;			// no longer the first time through
	    continue;
	}
	fseeko(af->aseg,segsize,SEEK_CUR); // seek to next segment
    } while(1);
    return -1;				// couldn't find segment
}





/* get_segment:
 * Get a segment. If it is compressed, uncompress it.
 * *data will point to the uncompressed segment.
 * if shouldfree is set, then data should be freed.
 * Return 0 if success, -1 if fail.
 */

int af_get_image_segment(AFFILE *af,int segnum,
			 unsigned char **data,size_t *bytes)
{
    char segname[256];
    unsigned long arg=0;
    unsigned char *segdata;
    size_t segdata_len;

    sprintf(segname,AF_SEG,segnum);
    if(af_get_seg(af,segname,&arg,&segdata,&segdata_len)){
	return -1;			// couldn't get it
    }
    
    /* See if we need to uncompress it */
    if((arg & AF_SEG_COMPRESSED)==0){
	/* It's not compressed; just return the raw segment. */
	*data = segdata;
	*bytes = segdata_len;
	return 0;
    }

    /* Need to uncompress it */
    *data = (unsigned char *)malloc(af->image_datasegsize);
    *bytes = af->image_datasegsize;

    uLongf uncompressed_bytes=af->image_datasegsize;

    int r = uncompress(*data,&uncompressed_bytes,(const Bytef *)segdata,segdata_len);
    switch(r){
    case Z_OK:
	*bytes = uncompressed_bytes;
	free(segdata);		// don't need this one anymore
	return 0;
    case Z_MEM_ERROR:
	err(1,"Z_MEM_ERROR decompressing segment %d",segnum);
    case Z_DATA_ERROR:
	err(2,"Z_DATA_ERROR decompressing segment %d",segnum);
    default:
	err(3,"uncompress returned an invalid value in get_segment");
    }
    return -1;				// should never get here!
}


unsigned long long af_decode_q(unsigned char buf[8])
{
    struct aff_quad *q = (struct aff_quad *)buf;		// point q to buf.

    assert(sizeof(*q)==8);		// be sure!
    return (((unsigned long long)ntohl(q->high)) << 32) +
	((unsigned long long)ntohl(q->low));
}

int af_get_segq(AFFILE *af,const char *name,unsigned long long *aff_quad)
{
    unsigned char *buf;
    size_t  bufsize;

    if(af_get_seg(af,name,0,&buf,&bufsize)){
	return -1;			// couldn't get it...
    }
    if(bufsize!=sizeof(struct aff_quad)){		// make sure size is good.
	free(buf);
	return -1;
    }
	
    *aff_quad = af_decode_q(buf);
    free(buf);				// free allocated stuff.
    return 0;
}

/* af_get_segsizes:
 * Get the segment sizes if they are set in the file.
 */
void af_get_segsizes(AFFILE *af)
{
    af_get_seg(af,AF_SEGSIZE,&af->image_datasegsize,0,0);
    af_get_segq(af,AF_IMAGESIZE,&af->imagesize);
    af->imagesize_orig = af->imagesize;
}


/* af_demand_segsizes:
 * Get the segment sizes and fail if we can't get them...
 */
void af_demand_segsizes(AFFILE *af)
{
    if(af->image_datasegsize==0){
	af_get_segsizes(af);
    }
    if(af->image_datasegsize==0){
	errx(1,"af_demand_segsizes: af->image_datasegsize not set. Aborting.\n");
    }
}


/****************************************************************
 *** Update functions
 ****************************************************************/
int af_rewindi(AFFILE *af)
{
    fseeko(af->aseg,sizeof(struct af_head),SEEK_SET); // go to the beginning
    return 0;
}

/* af_backspace:
 * moves back one segment in the aff file
 */
int af_backspace(AFFILE *af)
{
    struct af_segment_tail segt;

    aff_ull start = ftello(af->aseg);
    aff_ull pos_tail = start - sizeof(segt); // backspace to read the tail
    fseeko(af->aseg,pos_tail,SEEK_SET);
    if(fread(&segt,sizeof(segt),1,af->aseg)!=1){
	fseeko(af->aseg,start,SEEK_SET); // put it back
	return -1;			// can't read segt?
    }
    /* Now I know how long the segment was. Compute where it started */
    aff_ull seg_start = start - ntohl(segt.segment_len);
    fseeko(af->aseg,seg_start,SEEK_SET);
    return 0;
}

int af_enable_writing(AFFILE *af,int flag)
{
    int old_flag = af->writing;
    af->writing = flag;
    af->compression_callback = 0;	
    return old_flag;
}

int af_enable_compression(AFFILE *af,int flag,void (*cb)(int,int))
{
    int old_flag = af->no_compress;
    af->no_compress = (flag) ? 0 : 1;
    af->compression_callback = cb;
    return old_flag;
}

const char *af_filename(AFFILE *af)
{
    return af->fname;
}

const unsigned char *af_badflag(AFFILE *af)
{
    return af->badflag;
}


/*
 * af_update_segv:
 * Update the given named segment with the new value.
 * if "append" is false, don't append if it can't be found. 
 * if "append" is true and the named segment can't be found, append it.
 */

int af_update_seg(AFFILE *af, const char *name,
		    unsigned long arg,const void *value,unsigned int vallen,int append)
{
    char   next_segment_name[AF_MAX_NAME_LEN];
    size_t next_segsize = 0;
    size_t next_datasize = 0;

    /* if we are updating with a different size,
     * remember the location and size of the AF_IGNORE segment that
     * has the smallest size that is >= strlen(name)+vallen
     */
    size_t size_needed = strlen(name)+vallen+af_ignore_overhead();
    size_t size_closest = 0;
    aff_ull         loc_closest = 0;
       
#ifdef DEBUG
    fprintf(stderr,"af_update_seg(name=%s,arg=%lu,vallen=%u,append=%d)\n",name,arg,vallen,append);
#endif

    /* Is the last segment an ignore? */

    af_rewindi(af);			// start at the beginning

    while(af_probe_next_seg(af,next_segment_name,sizeof(next_segment_name),
			    0,&next_datasize,&next_segsize,1)==0){
	aff_ull next_segment_loc = ftello(af->aseg);

#ifdef DEBUG
	fprintf(stderr,"  next_segment_name=%s next_datasize=%d "
		"next_segsize=%d next_segment_loc=%qd\n",
		next_segment_name,
		next_datasize,
		next_segsize,next_segment_loc);
#endif

	if(strcmp(next_segment_name,name)==0){	// found the segment!

	    /* Okay. Is there room to fit it here? */
	    if(next_datasize == vallen){
		/* We can just write in place! */
		int r = af_put_seg(af,name,arg,value,vallen);
		return r;
	    }

	    /* Invalidate this segment and continue scanning */
	    
	    aff_ull loc_invalidated = next_segment_loc; // remember where this is

	    af_write_ignore(af,next_datasize+strlen(name));

	    aff_ull loc_before_seek = ftello(af->aseg); // before the seek

	    fseeko(af->aseg,(unsigned long long)0,SEEK_END);              // go to the end of the file

	    aff_ull pos_eof = ftello(af->aseg);


	    if(pos_eof == loc_before_seek){
		/* Apparently I didn't move at all with the seek.
		 * If the segment being written is larger than the segment being replaced,
		 * we can just backspace.
		 */
		if(next_segsize < vallen){
		    //fprintf(stderr,"  *** seeking to %qd\n",loc_invalidated);
		    fseeko(af->aseg,loc_invalidated,SEEK_SET);
		}
	    }

	    append = 1;			// be sure to add the data to the end!
	    break;			// and exit this loop
	}

	/* If this is an AF_IGNORE, see if it is a close match */
	if(next_segment_name[0]==AF_IGNORE[0]){
	    if(next_datasize>=size_needed && ((next_datasize<size_closest || size_closest==0))){
		size_closest = next_datasize;
		loc_closest  = next_segment_loc;
	    }
	}
	fseeko(af->aseg,next_segsize,SEEK_CUR); // skip this segment
    }
    if(append){				// if appending, just write to the end
	/* See if there is a place where this would fit. */
	//fprintf(stderr,"size_closest=%d\n",size_closest);
	if(size_closest>0){
	    /* Yes. Put it here and put a new AF_IGNORE in the space left-over
	     * TODO: If the following space is also an AF_IGNORE, then combine the two.
	     */
#ifdef DEBUG
	    fprintf(stderr,"size_closest=%d\n",size_closest);
	    fprintf(stderr,"loc_closest=%qd\n",loc_closest);
	    fprintf(stderr,"*** Squeezing it in at %qd. name=%s. vallen=%d\n",loc_closest,name,vallen);
#endif

	    fseeko(af->aseg,loc_closest,SEEK_SET); // move to the location
	    af_put_seg(af,name,arg,value,vallen); // write the new segment

	    int newsize = size_closest - vallen - af_ignore_overhead() - strlen(name);

	    af_write_ignore(af,newsize); // write the smaller ignore
	    return 0;
	}
	return af_put_seg(af,name,arg,value,vallen);
    }
    return -2;				// couldn't find segment
}


/* af_update_segq:
 * Update the named aff_quad-byte value.
 */

int af_update_segq(AFFILE *af, const char *name, unsigned long long value,int append)
{
    struct aff_quad  q;
    q.low  = htonl(value & 0xffffffff);
    q.high = htonl(value >> 32);
    
    return af_update_seg(af,name,0,&q,8,append);
}


/****************************************************************
 *** AFF creation functions
 ****************************************************************/

int af_put_badflag(AFFILE *af)
{
    RAND_pseudo_bytes(af->badflag,sizeof(af->badflag));
    strcpy((char *)af->badflag,"BAD SECTOR");

    if(af_put_seg(af,AF_BADFLAG,0,af->badflag,sizeof(af->badflag))) return -1;
    if(af_put_seg(af,AF_BADSECTORS,0,0,0)) return -1;
    return 0;
}


/*
 * af_set_sizes:
 * Sets the segment and image size.
 * Doesn't accept new datasegsize ones if they are already set...
 */
void af_set_sizes(AFFILE *af,long data_segsize,unsigned long long imagesize)
{
    if(af->image_datasegsize==0 && data_segsize!=0){
	af_put_seg(af,AF_SEGSIZE,data_segsize,0,0);
	af->image_datasegsize	= data_segsize;
    }

    af_put_segq(af,AF_IMAGESIZE,imagesize);
    af->imagesize		= imagesize;
    af->imagesize_orig		= imagesize;
}







