#ifndef _AFFLIB_H_
#define _AFFLIB_H_ 1

#define AFF_VERSION "1.1"
#define AFF_SECTOR_SIZE 512

/* Horrible lossage stuff for largefile support under Linux */


#ifdef linux
#define _LARGEFILE_SOURCE 1
#define _FILE_OFFSET_BITS 64
#endif



#include <stdio.h>

typedef struct _AFFILE {
    /* For Raw files */
    FILE *raw;				// if it is a raw file

    /* For AFF Segment Files */
    FILE *aseg;

    /* Filename of file... */
    char *fname;			// 

    /* Image information */
    unsigned long image_datasegsize;  // the size of image data segments in this file
    unsigned long long imagesize;	// how big is the disk image
    unsigned char badflag[AFF_SECTOR_SIZE];		// bad sector flag

    /* Implement a stream abstraction */
    int     segnum;			// -1 means no segment loaded
    unsigned char *segbuf;		// where the data is
    size_t   segbytes;			// number of bytes that were loaded from disk
    unsigned long long pos;		// location in stream

    /* additional support for writing. */
    unsigned int dirty:1;		// is buffer dirty?
    unsigned int writing:1;		// is file open for writing?
    unsigned int segs_written;		// segments written
    unsigned int csegs_written;		// compressed segments written
    unsigned long long imagesize_orig;	// see if it was changed...
    unsigned int no_compress:1;		// don't try to compress

    void (*compression_callback)(int bytes_uncompressed,
				 int bytes_compressed);
    // called at start and end of compression.

    int last_command;			// 1 for probe, 2 for put
    int debug;				// for debugging, of course
} AFFILE;

/* The header for an AFF file. All binary numbers are stored in network byte order. */
struct af_head {
    char header[8];			// "AFF10\r\n\000"
    /* segments follow */
};

#define AF_HEADER "AFF10\r\n\000"

struct af_segment_head {
    char magic[4];			// "AFF\000"
    unsigned long name_len:32;		// length of segment name
    unsigned long data_len:32;		// length of segment data, if any
    unsigned long flag:32;		// argument for name;
    /* name follows, then data */
};

struct af_segment_tail {
    char magic[4];			// "ATT\000"
    unsigned long segment_len:32;      // includes head, tail, name & length
};

#define AF_SEGHEAD "AFF\000"
#define AF_SEGTAIL "ATT\000"


typedef unsigned long long aff_ull ;

struct aff_quad {
    unsigned long low:32;
    unsigned long high:32;
};



struct af_dir {
    int len;				// length of directory entry, including name and header
    struct aff_quad offset;		// offset, stored as an aff_quad
    char name[0];			// name of the entry
};

/* Utility Functions */

#ifdef __cplusplus
extern "C" {
#endif
#ifdef __never_defined__
}
#endif

/****************************************************************
 ***
 *** Intended user AFF interface
 ***
 ****************************************************************/

/* af_file stream functions */
int	af_identify(const char *filename);  // 1 = yes, 0 = no, -1 = error (no file)


AFFILE *af_open(const char *filename,int flags,int mode);
int	af_close(AFFILE *af);
const char *af_filename(AFFILE *af);	// returns the filename of an open stream.

/* Stream-based access */

int	af_eof(AFFILE *af);
int	af_read(AFFILE *af,unsigned char *buf,int count);
aff_ull af_seek(AFFILE *af,aff_ull pos,int whence); // returns new position
aff_ull af_tell(AFFILE *af);

/* Additional routines for writing */
int	af_enable_writing(AFFILE *af,int flag);	// set true to enable writing; returns old value
int	af_enable_compression(AFFILE *af,int flag,
			      void (*cb)(int,int)); // set/gunset compression for writing
int	af_write(AFFILE *af,unsigned char *buf,int count);



/* Metadata access */

/* get functions:
 * get the named segment.
 * If arg!=0, set *arg to be the segment's flag.
 * If data!=0, allocate space for the data and set *data to the buffer.
 * (Free with free()).
 * If datalen!=0, set *datalen to be the length of the segment.
 */

int	af_get_seg(AFFILE *af,const char *name,unsigned long *flag,
		    unsigned char **data,size_t *datalen);


int	af_get_segq(AFFILE *af,const char *name,unsigned long long *quad);


/* update functions
 * af_update_seg() should be your primary routine for writing new values.
 * if append==0, then the value will not be written if the segment doesn't
 * already exist. If append!=0, then the segment will be appended to the end
 * of the file if it is not there already.
 */

int	af_update_seg(AFFILE *af,const char *name,unsigned long arg,
		       const void *value,unsigned int vallen,int append); 

int	af_update_segq(AFFILE *af,const char *name,unsigned long long quad,int append);


/****************************************************************
 *** AF segment names that you might be interested in...
 ****************************************************************/


#define AF_IGNORE       ""		// ignore segments with 0-length name
#define AF_SEGSIZE	"segsize"	// segment data size
#define AF_IMAGESIZE	"imagesize"	// overall size of the image; 8-bytes
#define AF_BADSECTORS	"badsectors"	// number of bad sectors
#define AF_BADFLAG      "badflag"	// data used to mark a bad sector
#define AF_SEG          "seg%d"		// arg=1 means it is compressed with zlib
#define AF_BLANKSECTORS "blanksectors"	// all NULs; 8-bytes

/* AFF Flags */

#define AF_SEG_COMPRESSED 0x0001
#define AF_SEG_QUADWORD   0x0002	// interpert this as a quadword
#define AF_MD5  "md5"			// stores image md5
#define AF_SHA1 "sha1"			// stores image sha1

/* segment names: imaging */
#define AF_DATE_ACQUIRED	"date_acquired" // YYYY-MM-DD HH:MM:SS TZT
#define AF_ISO_COUNTRY_ACQUIRED "iso_country_acquired" // ISO country code
#define AF_CASE_NUM		"case_num"      // case number
#define AF_IMAGING_NOTES	"imaging_notes" // notes made while imaging
#define AF_ACCESSION_GID	"accession_gid" // 128-bit unique accession number
#define AF_IMAGING_DEVICE	"imaging_device" // device used to do the imaging


/* segment names: device hardware */

#define AF_DEVICE_MANUFACTURER  "device_manufacturer"
#define AF_DEVICE_MODEL		"device_model"	// string for ident from drive
#define AF_DEVICE_SN		"device_sn"	// string of drive capabilities
#define AF_DEVICE_FIRMWARE	"device_firmware"	// string of drive capabilities
#define AF_DEVICE_SOURCE        "device_source" // string
#define AF_CYLINDERS		"cylinders" // quad with # cylinders
#define AF_HEADS		"heads"	// quad with # heads
#define AF_SECTORS_PER_TRACK	"sectors/track"	// quad with # sectors/track
#define AF_LBA_SIZE		"lbasize"
#define AF_HPA_PRESENT          "hpa_present"   // flag = 1 or 0
#define AF_HCO_PRESENT          "dco_present"   // flag = 1 or 0
#define AF_LOCATION_IN_COMPUTER "location_in_computer" // text, where it was found
#define AF_DEVICE_CAPABILITIES	"device_capabilities" // string; human-readable

#define AF_MAX_NAME_LEN 64		// segment names should not be larger than this

/****************************************************************
 *** Internal implementation details below.
 ****************************************************************/


const char *af_hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int add_spaces);

/* Low-level functions for putting
 * af_put_segv --- put a segment with an argument (also useful for putting numbers)
 * af_put_segq --- put a quadword, useful for putting 8-byte values.
 */

int	af_put_segq(AFFILE *af,const char *name,unsigned long long value);

int	af_put_seg(AFFILE *af,const char *name,unsigned long arg,
		   const void *value,unsigned int vallen); 
int	af_del_seg(AFFILE *af,const char *name); // complete delete of first name
                                                 // returns 0 if success, -1 if seg not found

/* Navigating within the AFFILE */
/* probe the next segment.
 * Returns: 0 if success
 *          -1 if error
 *          -2 if segname_len was not large enough to hold segname
 *         - segname - the name of the next segment.
 *         - segsize - number of bytes the entire segment is.
 *      
 * doesn't change af->aseg pointer if do_rewind is true, otherwise leaves stream
 *           positioned ready to read the data
 */

int	af_probe_next_seg(AFFILE *af,char *segname,size_t segname_len,
			   unsigned long *arg,
			   size_t *datasize,
			   size_t *segsize,
			   int do_rewind);

int	af_find_seg(AFFILE *af,const char *segname,
		    unsigned long *arg,
		    size_t *datasize,
		    size_t *segsize);	
		    


int	af_put_badflag(AFFILE *af);	// creates a badflag and puts it


/* get functions */
int	af_get_next_seg(AFFILE *af,char *segname,size_t segname_len,
			 unsigned long *arg,
			 unsigned char **data,
			 size_t *datalen);
int	af_get_image_segment(AFFILE *af,int segnum,unsigned char **data,size_t *bytes);

unsigned long long af_decode_q(unsigned char buf[8]); // return buf[8] into an unsigned quad



int	af_rewindi(AFFILE *af);		// go to the beginning of the AF file; doesn't move
                                        // the file pointer; use af_seek(af,0,0);
int	af_backspace(AFFILE *af);	// back up one segment


int	af_purge(AFFILE *af);		// throw out the current data segment in memory
const unsigned char *af_badflag(AFFILE *af); // return the flag used to identify bad blocks
int	af_ignore_overhead();


void	af_get_segsizes(AFFILE *af);	// sets up values if we can get them.
void	af_demand_segsizes(AFFILE *af);	// abort if we can't get them

/* aff creation functions */
void	af_set_sizes(AFFILE *af,long segsize,unsigned long long imagesize);

/* Return parameters about a raw device; have a value as 0 if it doesn't matter. */
int	af_figure_media(int fd,int *sector_size,unsigned long long *total_sectors,int *maxreadblocks);



#ifdef __never_defined__
{
#endif
#ifdef __cplusplus
}
#endif
#endif
