/*
 * The AFFLIB data stream interface.
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */



/* Be sure we have 64-bit support */
#ifndef LARGEFILE_SOURCE
#define LARGEFILE_SOURCE 1
#endif

#ifndef _FILE_OFFSET_BITS
#define FILE_OFFSET_BITS 64
#endif


#include "afflib.h"
#include <stdio.h>
#include <err.h>
#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <openssl/rand.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>


#ifndef O_BINARY
#define O_BINARY 0			// for Windows compatability
#endif



AFFILE *af_open(const char *filename,int flags,int mode)
{
    /* make sure things were compiled properly */
    assert(sizeof(struct af_head)==8);
    assert(sizeof(struct af_segment_head)==16);
    assert(sizeof(struct af_segment_tail)==8);

    if(flags & O_WRONLY) err(1,"af_open: flag O_WRONLY not supported");

    int exist = (access(filename,R_OK) == 0);	// does the file exist?
    int fd = open(filename,flags | O_BINARY,mode);
    if(fd<0) return 0;			// couldn't open

    AFFILE *af = (AFFILE *)calloc(sizeof(AFFILE),1);

    af->fname	= strdup(filename);	// remember file name
    af->segnum  = -1;			// no segment loaded

    int creating = flags & O_TRUNC ? 1 : 0; // are we creating? 
    struct stat sb;


    if(exist && !creating && af_identify(filename)!=1){
	/* It's not an AFF file. If the file is more than 0 bytes long,
	 * then pass through to raw routines...
	 */
	af->raw=fdopen(fd,"rb");
	
	if(stat(filename,&sb)){
	    err(1,"stat(%s): ",filename);
	}
	af->imagesize = sb.st_size;
	return af;
    }
	
    /* If we are creating, we must be creating an AFF... */
    char strflag[8];
    strcpy(strflag,"r");		// we have to be able to read
    if(flags & O_RDWR) 	strcpy(strflag,"w+");

    af->aseg = fdopen(fd,strflag);
    if(!af->aseg) err(1,"fdopen(%d,%s)",fd,strflag);

    /* If file is empty, then make it an AFF file! */
    if(fstat(fd,&sb)){
	err(1,"fstat(%s): ",filename);
    }
    if(sb.st_size==0){
	fwrite(AF_HEADER,1,8,af->aseg);	// writes the header 
	af_put_badflag(af);		// writes the flag for bad blocks
	return af;
    }
    
    /* We are opening an existing file. Verify once more than it is an AFF file
     * and skip past the header...*/

    /* Oh, we are reading. See if it is an AF header */
    char buf[8];
    if(fread(buf,sizeof(buf),1,af->aseg)!=1){
	/* Hm. End of file. That shouldn't happen here. */
	err(1,"af_open: couldn't read AFF header on existing file?");
    }

    if(strcmp(buf,AF_HEADER)!=0){
	buf[7] = 0;
	errx(1,"%s is not an AFF file (header=%s)\n",filename,buf);
    }
    af_get_segsizes(af);		// get the segment sizes if we can...
    return af;				// everything must be okay.
}

/* Close the image */
int af_close(AFFILE *af)
{
    if(!af) return -1;			// invalid handle
    af_purge(af);			// get rid of segbuf
    if(af->writing){
	if(af->imagesize != af->imagesize_orig){
	    af_update_segq(af,AF_IMAGESIZE,af->imagesize,1);
	}
    }
    if(af->raw)   fclose(af->raw);
    if(af->aseg)  fclose(af->aseg);
    if(af->fname) free(af->fname);
    memset(af,0,sizeof(*af));
    free(af);
    return 0;
}


/* Return if we are at the end of the file */


int af_eof(AFFILE *af)
{
    if(af->raw) return feof(af->raw);
    if(af->image_datasegsize==0) af_demand_segsizes(af);
    return af->pos >= af->imagesize;
}

/* Throw out the current segment */
int af_purge(AFFILE *af)
{
    /* If the segment is dirty, we need to write it back
     * (and, if there is already a segment, we need to invalidate it.)
     */
    if(af->dirty){
	 char vbuf[32];
	 sprintf(vbuf,AF_SEG,af->segnum);

	 /* Try to compress the data */
	 uLongf destLen = af->image_datasegsize;
	 unsigned char *cdata = 0;
	 int cres = -1;			// compression result; default is that we failed

	 cdata = (unsigned char *)malloc(af->image_datasegsize);
	 if(af->no_compress == 0 && cdata!=0){		// try to compress
	     if(af->compression_callback){
		 (*af->compression_callback)(0,0); // compression starting
	     }
	     cres = compress((Bytef *)cdata,&destLen,(Bytef *)af->segbuf,af->image_datasegsize);
	     if(af->compression_callback){
		 (*af->compression_callback)(af->image_datasegsize,
					     destLen);
		 // compression done
	     }
	 }

	 if(cres==0 && destLen < af->image_datasegsize){
	     /* Prepare to write out the compressed segment with compression */
	     if(af_update_seg(af,vbuf,AF_SEG_COMPRESSED,cdata,destLen,1)) return -1;
	     af->csegs_written++;
	 }
	 else {
	     /* Prepare to write out the segment without compression */
	     if(af_update_seg(af,vbuf,0,af->segbuf,af->image_datasegsize,1)) return -1;
	     af->segs_written++;
	 }

	 af->dirty = 0;
	 if(cdata){
	     free(cdata);
	     cdata = 0;
	 }
    }

    if(af->segbuf){
	free(af->segbuf);
	af->segbuf = 0;
	af->segnum = -1;
    }
    return 0;
}


/* Read from the affile */
int af_read(AFFILE *af,unsigned char *buf,int count)
{
    if(af->raw){	/* passthrough? */
	return fread(buf,1,count,af->raw);
    }


    af_demand_segsizes(af);    /* We need to have a datasegsize to continue */

    int total = 0;
    while(count>0){
	/* If no segment is loaded, load the current segment */
	if(af->segbuf==0){
	    af->segnum = af->pos / af->image_datasegsize;	// will be the segment we want
	    if(af_get_image_segment(af,af->segnum,&af->segbuf, &af->segbytes)){
		break;			// no more to get
	    }
	}
	// Compute how many bytes can be copied...
	// where we were reading from
	unsigned int seg_offset = af->pos - af->segnum * af->image_datasegsize; 
	unsigned int seg_left   = af->segbytes - seg_offset; // number we can get out
	unsigned int bytes_to_read = count;

	if(bytes_to_read > seg_left) bytes_to_read = seg_left;
	if(bytes_to_read > af->imagesize - af->pos) bytes_to_read = af->imagesize - af->pos;

	assert(bytes_to_read >= 0);	// 
	if(bytes_to_read==0) return total; // that's all we could get

	/* Copy out the bytes for the user */
	memcpy(buf,af->segbuf+seg_offset,bytes_to_read); // copy out
	buf     += bytes_to_read;
	af->pos += bytes_to_read;
	count   -= bytes_to_read;
	total   += bytes_to_read;

	/* If we read out all of the bytes that were left in the segment,
	 * then we are at the end of the segment, so throw it away.
	 */
	if(seg_left == bytes_to_read){	
	    af_purge(af);
	}
    }
    /* We have copied all of the user's requested data, so return */
    return total;
}

int af_write(AFFILE *af,unsigned char *buf,int count)
{
    if(!af->writing){
	errno = EPERM;			// operation not permitted
	return -1;			// not opened for writing
    }
    af_demand_segsizes(af);

    int total = 0;
    while(count>0){
	/* If no segment is loaded, load the current segment */
	if(af->segbuf==0){
	    af->segnum = af->pos / af->image_datasegsize;	// will be the segment we want
	    if(af_get_image_segment(af,af->segnum,&af->segbuf, &af->segbytes)){
		/* Create a new segment! */
		af->segbuf   = (unsigned char *)malloc(af->image_datasegsize);
		memset((char *)af->segbuf,'E',af->image_datasegsize);
		af->segbytes = 0;

		/* Fill the segment with the bad block flag... */
		unsigned char *cc;
		int i=0;
		for(cc=af->segbuf;cc<af->segbuf+af->image_datasegsize;cc++){
		    *cc = af->badflag[i++];
		    if(i>AFF_SECTOR_SIZE) i=0;
		}
	    }
	}
	unsigned int seg_offset = af->pos - af->segnum * af->image_datasegsize; // where we were writing to
	unsigned int seg_left   = af->image_datasegsize - seg_offset; // number we can write into
	unsigned int bytes_to_write = count;

	if(bytes_to_write > seg_left) bytes_to_write = seg_left;

	assert(bytes_to_write >= 0);	// 
	if(bytes_to_write==0) return total; // that's all we could get

	/* Copy out the bytes for the user */
	memcpy(af->segbuf+seg_offset,buf,bytes_to_write); // copy out

	if(af->segbytes < seg_offset+bytes_to_write){
	    af->segbytes = seg_offset+bytes_to_write; // it has been extended.
	}

	buf     += bytes_to_write;
	af->pos += bytes_to_write;
	count   -= bytes_to_write;
	total   += bytes_to_write;
	af->dirty = 1;

	/* If we wrote out all of the bytes that were left in the segment,
	 * then we are at the end of the segment, write it back...
	 */
	if(seg_left == bytes_to_write){	
	    if(af_purge(af)) return -1;
	}

	/* If we have written more than the image size, update the image size */
	if(af->pos > af->imagesize){
	    af->imagesize = af->pos;
	}

    }
    /* We have copied all of the user's requested data, so return */
    return total;
}



/* Seek in the virtual file */
aff_ull af_seek(AFFILE *af,aff_ull pos,int whence)
{
    unsigned long long new_pos;

    if(af->raw) return fseeko(af->raw,pos,whence);
    af_demand_segsizes(af);

    switch(whence){
    case SEEK_SET:
	new_pos = pos;
	break;
    case SEEK_CUR:
	new_pos += pos;
	break;
    case SEEK_END:
	new_pos = af->imagesize - pos;
	break;
    }
    if(new_pos < 0) new_pos = 0;	   // new change
    if(af->pos == new_pos) return af->pos; // no change
    af->pos = new_pos;			// set the new position

    /* If the correct segment is not loaded, purge the segment */
    int new_seg = af->pos / af->image_datasegsize;
    if(new_seg != af->segnum){
	af_purge(af);
    }

    return af->pos;
}

aff_ull  af_tell(AFFILE *af)
{
    if(af->raw) return ftello(af->raw);
    return af->pos;
}


