/*
 * ainfo.cpp:
 *
 * print information about an aff file
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>
#include <curses.h>
#include <term.h>

char *progname = "acat";

int opt_validate = 0;
int opt_info = 1;
int opt_all  = 0;
int opt_wide = 0;
int cols = 0;
char **opt_j = 0;
int opt_j_count = 0;

void usage()
{
    printf("usage: %s [options] infile\n",progname);
    printf("   -a = print ALL segments (normally data segments are suppressed)\n");
    printf("   -v = validate the file's md5 and sha-1 codes\n");
    printf("   -i = print informaton about each segment %s\n",opt_info ? "(DEFAULT)" : "");
    printf("   -I = don't print info about each segment %s\n",opt_info ? "" : "(DEFAULT)");
    printf("   -w = wide output; print more than 1 line if necessary.\n");
    printf("   -j blockname =   Just print <blockname> for the file.\n");
    printf("                    (may be repeated; implies -I)\n");
    exit(0);
}



void validate(const char *infile)
{
    printf("\nValidating hash codes...\n");

    /* Set up the hash machinery */
    MD5_CTX md5;
    MD5_Init(&md5);

    SHA_CTX sha;
    SHA1_Init(&sha);
    
    AFFILE *af = af_open(infile,O_RDONLY,0);
    unsigned long long total_bytes = 0;
    while(!af_eof(af)){
	unsigned char buf[65536];		// a decent size
	size_t bytes = af_read(af,buf,sizeof(buf));
	total_bytes += bytes;
	MD5_Update(&md5,buf,bytes);
	SHA1_Update(&sha,buf,bytes);
    }
    
    /* Finish the hash calculations and write to the db */
    unsigned char *md5_data;
    size_t md5len;
    unsigned char ibuf[20];
    char buf[256];
    unsigned long arg;
    
    MD5_Final(ibuf,&md5);
    printf("md5: %s ",af_hexbuf(buf,sizeof(buf),ibuf,16,1));
    if(af_get_seg(af,AF_MD5,&arg,&md5_data,&md5len)==0){
	if(md5len==16 && !memcmp((const char *)md5_data,(const char *)ibuf,16)){
	    printf(" MATCH\n");
	}
	else {
	    printf(" NO MATCH!\n");
	}
	free(md5_data);
    }
    else {
	printf("(no MD5 in AFF file)\n");
    }
    
    
    unsigned char *sha1_data;
    size_t sha1len;

    SHA1_Final(ibuf,&sha);
    printf("sha1: %s ",af_hexbuf(buf,sizeof(buf),ibuf,20,1));
    if(af_get_seg(af,AF_SHA1,&arg,&sha1_data,&sha1len)==0){
	if(sha1len==20 && !memcmp((const char *)sha1_data,(const char *)ibuf,20)){
	    printf(" MATCH\n");
	}
	else {
	    printf(" NO MATCH!\n");
	}
	free(sha1_data);
    }
    else {
	printf("(no SHA1 in AFF file)\n");
    }
    
    struct stat sb;
    if(fstat(fileno(af->aseg),&sb)){
	err(1,"fstat");
    }
    
    printf("AFF file is %qd bytes. Actual compression ratio with overhead: %4.1f%%\n",
	   (long long int)sb.st_size,
	   ((float)total_bytes-sb.st_size)*100.0/total_bytes);
    af_close(af);
}

/* Return true if segname is in the optj list */
bool in_opt_j_list(char *segname)
{
    for(int i=0;i<opt_j_count;i++){
	if(strcmp(segname,opt_j[i])==0) return true;
    }
    return false;
}



int get_info(const char *infile)
{
    char *outline = (char *)alloca(4096);
    unsigned long arg;
    unsigned long total_segs = 0;
    unsigned long total_data_segs = 0;

    AFFILE *af = af_open(infile,O_RDONLY,0);

    printf("\n%s\n",af_filename(af));
    if(opt_all==0){
	printf("[skipping data segments]\n");
    }
    printf("                       file                data        \n");
    printf("Segment              offset      arg     length    data\n");
    printf("=======             =======  =======     ======    =====\n");

    af_rewindi(af);			// start at the beginning
    while(1){
	char segname[AF_MAX_NAME_LEN];
	char segname2[AF_MAX_NAME_LEN];
	size_t datalen = 0;

	aff_ull loc = ftello(af->aseg);
	if(af_get_next_seg(af,segname,sizeof(segname),&arg,0,&datalen)!=0){
	    printf("==EOF==\n");
	    break;
	}

	total_segs++;

	/* See if it is a data segment */
	int dataseg_num = 0;
	if(sscanf(segname,AF_SEG,&dataseg_num)==1){
	    total_data_segs++;
	    if(opt_all==0) continue;
	}

	outline[0] = 0;
	sprintf(outline,"%-15s    %8qd %8ld   %8ld    ",
		segname[0] ? segname : "  <ignore>" ,
		loc,arg,(long)datalen);

	/* Get the data */
	af_backspace(af);
	unsigned char *data = 0;
	if(af_get_next_seg(af,segname2,sizeof(segname2),&arg,&data,&datalen)!=0){
	    warn("af_get_next_seg #2 failed");
	    break;
	}


	/* Special handling of quadwords? */
	if(arg & AF_SEG_QUADWORD || strcmp(segname,AF_IMAGESIZE)==0){
	    /* Print it as a 64-bit value.
	     * The strcmp is there because early AF_IMAGESIZE segs didn't set
	     * AF_SEG_QUADWORD...
	     */
	    printf("%s= %qd (64-bit value)\n",outline,af_decode_q(data));
	    free(data);
	    continue;
	}


	/* See if I need to truncate */
	int dots = 0;
	if(datalen>32){		// don't bother showing more than first 32 bytes
	    dots = 1;
	    datalen = 32;
	}
	    
	char buf[80];
	char *cc = outline + strlen(outline);
	sprintf(cc,"%s%s",
		af_hexbuf(buf,sizeof(buf),data,datalen,1),
		dots ? "..." : "");
	
	/* Now print the results... */
	if(!opt_wide){
	    if(strlen(outline)>4){
		outline[cols-4] = '.';
		outline[cols-3] = '.';
		outline[cols-2] = '.';
		outline[cols-1] = '\000';
	    }
	}
	puts(outline);
	free(data);
    }
    printf("\n");
    printf("Data segments:  %8lu\n",total_data_segs);
    printf("Total segments: %8lu\n",total_segs);
    af_close(af);
    return 0;
}
	 




int main(int argc,char **argv)
{
    int bflag, ch;
    const char *infile;

    /* Figure out how many cols the screen has... */
    setupterm((char *)0,1,(int *)0);
    cols = tgetnum("co");

    progname = argv[0];
    bflag = 0;
    while ((ch = getopt(argc, argv, "ah?viIwj:")) != -1) {
	switch (ch) {
	case 'a':
	    opt_all++;
	    break;
	case 'h':
	case '?':
	default:
	    usage();
	    break;
	case 'i':
	    opt_info++;
	    break;
	case 'I':
	    opt_info=0;
	    break;
	case 'v':
	    opt_validate++;
	    break;
	case 'w':
	    opt_wide++;
	    break;
	case 'j':
	    if(opt_j==0) opt_j = (char **)malloc(0);
	    opt_j_count++;
	    opt_j = (char **)realloc(opt_j,sizeof(char *)*opt_j_count);
	    opt_j[opt_j_count-1] = strdup(optarg); // make a copy
	    
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }


    /* Loop through all of the files */
    while(*argv){
	infile = *argv++;		// get the file
	argc--;				// decrement argument counter
	 
	switch(af_identify(infile)){
	case -1:
	    err(1,infile);		// can't read
	case 0:
	    fprintf(stderr,"%s is not an AFF file.\n",infile);
	    continue;
	}
	 
	if(opt_info)	 get_info(infile);
	if(opt_validate) validate(infile);
    }
    return 0;
}


