/*
 * atest.cpp:
 * test suite for the AFF Library.
 */

#include "afflib.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <assert.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <openssl/rand.h>


char *fmt = "This is format string %8d.\n";

/* Create the segment that we need */


#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif


int sequential_test()
{
    char buf[1024];

    printf("Sequential test...\n");
    AFFILE *af = af_open("test_sequential.aff",O_CREAT|O_RDWR|O_TRUNC,0666);
    if(!af) err(1,"af_open");
    af_enable_writing(af,1);
    af_set_sizes(af,1024,0);
    for(int i=0;i<1000;i++){
	printf("writing %d\n",i);
	sprintf(buf,fmt,i);
	if(af_write(af,(unsigned char *)buf,strlen(buf))!=(int)strlen(buf)){
	    fprintf(stderr,"Attempt to write buffer %d failed\n",i);
	    exit(1);
	}
    }
    af_close(af);
    return 0;
}

int reverse_test()
{
    char buf[1024];

    printf("Reverse write test...\n");
    AFFILE *af = af_open("test_reverse.aff",O_CREAT|O_RDWR|O_TRUNC,0666);
    if(!af) err(1,"af_open");
    af_set_sizes(af,1024,0);
    af_enable_writing(af,1);
    for(int i=999;i>=0;i--){
	sprintf(buf,fmt,i);
	af_seek(af,strlen(buf)*i,SEEK_SET);
	if(af_write(af,(unsigned char *)buf,strlen(buf))!=(int)strlen(buf)){
	    fprintf(stderr,"Attempt to write buffer %d failed\n",i);
	    exit(1);
	}
	printf("%d  ",i);
	fflush(stdout);
    }
    af_close(af);
    return 0;
}



int random_write_test()
{
    int count = 0;
    char buf[1024];


    /* Create the AFF file */
    sprintf(buf,fmt,"0");		// figure out how big fmt string is
    //int fmt_size = strlen(fmt);

    printf("Random write test...\n");

    //int seg_size = 1024;

    //printf("Writing %d byte records with a %d byte seg_size.\n",fmt_size,seg_size);

    AFFILE *af = af_open("test_random.aff",O_CREAT|O_RDWR|O_TRUNC,0666);
    if(!af) err(1,"af_open");
    af_set_sizes(af,1024,0);
    af_enable_writing(af,1);
    sprintf(buf,fmt,0);
    af_write(af,(unsigned char *)buf,strlen(buf)); // put out the first
    for(int i=0;i!=0;i = (i+3507) % 7001){
	sprintf(buf,fmt,i);
	af_seek(af,strlen(buf)*i,SEEK_SET);
	if(af_write(af,(unsigned char *)buf,strlen(buf))!=(int)strlen(buf)){
	    fprintf(stderr,"Attempt to write buffer %d failed\n",i);
	    exit(1);
	}
	printf("%d  (%d to go)\n",i,7001-count);
	fflush(stdout);
	count++;
    }
    af_close(af);
    printf("\ncount=%d\n",count);
    return 0;
}

int random_read_test(int total_bytes,int data_segment_size)
{
    printf("\n\n\nrandom read test. filesize=%d, segment_size=%d\n",total_bytes,data_segment_size);

    /* Create a regular file and an AFF file */

    printf("Creating file.img and file.aff, both with %d bytes of user data...\n",total_bytes);

    int    fd = open("file.img",O_CREAT|O_RDWR|O_TRUNC,0666);
    if(fd<0) err(1,"fopen");

    AFFILE *af = af_open("file.aff",O_CREAT|O_RDWR|O_TRUNC,0666);
    if(!af) err(1,"af_open");
    af_set_sizes(af,data_segment_size,0);
    af_enable_writing(af,1);
    /* Just write it out as one big write */

    unsigned char *buf = (unsigned char *)malloc(total_bytes);
    unsigned char *buf2 = (unsigned char *)malloc(total_bytes);
    RAND_pseudo_bytes(buf,total_bytes);
    if(write(fd,buf,total_bytes)!=total_bytes) err(1,"fwrite");
    if(af_write(af,buf,total_bytes)!=(int)total_bytes) err(1,"af_write");

    /* Now try lots of seeks and reads */
    for(int i=0;i<1000;i++){
	unsigned int loc = rand() % total_bytes;
	unsigned int len = rand() % total_bytes;
	memset(buf,0,total_bytes);
	memset(buf2,0,total_bytes);

	printf("#%d  reading %u bytes at %u\n",i,loc,len);

	unsigned long l1 = lseek(fd,loc,SEEK_SET);
	unsigned long l2 = af_seek(af,loc,SEEK_SET);


	if(l1!=l2){
	    err(1,"l1 (%lu) != l2 (%lu)",l1,l2);
	}

	int r1 = read(fd,buf,len);
	int r2 = af_read(af,buf2,len);

	if(r1!=r2){
	    err(1,"r1 (%d) != r2 (%d)",r1,r2);
	}
    }
    af_close(af);
    close(fd);
    return 0;
}


int main(int argc,char **argv)
{
    int do_sequential = 0;
    int do_reverse    = 0;
    int do_random_write_test = 0;
    int do_random_read_test  = 1;


    if(do_sequential){
	sequential_test();
    }
    if(do_reverse){
	reverse_test();
    }

    if(do_random_write_test){
	random_write_test();
    }

    if(do_random_read_test){
	for(int i=0;i<10;i++){
	    random_read_test(256*1024,rand() % 65536); // pick a random segment size
	}
    }
       return 0;
}
