/*
 * aimage.cpp:
 *
 * Image a physical drive connected through an IDE, SCSI, USB or Firewire interface
 * and write the results to either a raw or AFF file.
 *
 * Automatically calculate the MD5 & SHA-1 while image is made.
 * Handle read errors in an intelligent fashion. 
 * Display the results. 
 */

#include "afflib.h"
#include "aimage.h"

#include <time.h>
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <term.h>
#include <ctype.h>
#include <errno.h>

#include "imager.h"

#ifdef __FreeBSD__
#include <sys/param.h>
#include <sys/mount.h>
#endif

char cmd_attach[255]={0};		// to detach a device that was attached
char cmd_detach[255]={0};		// to detach a device that was attached
bool finishing = false;
int  compressing=0;

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel.
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */



char *progname = "aimage";

int  opt_datasegsize = 16*1024*1024;	// default seg size --- 16MB
int  opt_append = 0;
int  opt_quiet = 0;
int  opt_silent = 0;
int  opt_beeps  = 1;		// beep when done
int  opt_readsectors = 512;	// read in 256K chunks
int  opt_zap = 0;
char *opt_config=0;
int  opt_error = 0;
int  opt_retry = 5;
int  opt_reverse = 0;
char *infile = 0;
const char *outfile_raw =0 ;
const char *outfile_aff =0;
time_t  start_time = 0;
const char *logfile = 0;

int opt_skip = 0;			// what to skip on input
int opt_skip_sectors = 0;		// are we skipping sectors?
int opt_nocompress = 0;

char device_model[256] = {0};
char serial_number[256] = {0};
char firmware_revision[256] = {0};


void usage()
{
    putchar('\n');
    printf("usage: %s [options] INPUT --- image to a file\n",progname);
    printf("INPUT may be any of these:\n");
    printf("  A device (e.g. /dev/disk1).\n");
    printf("  - (or /dev/tty, for standard input)\n");
    
    //printf("  server:nnnn       Listen on TCP port nnnn\n");
    bold("\nGeneral Options:\n");
    printf("  --quiet, -q           -- No interactive statistics.\n");
    printf("  --silent, -Q          -- No output at all except for errors.\n");
    printf("  --readsectors=nn, -R nnnn,   -- set number of sectors to read at once (default %d)\n",
	                                      opt_readsectors);
    printf("  --skip=nn[s], -k nn   -- Skip nn bytes [or nns for sectors] in input file\n");
    printf("  --no_beeps, -B        -- Don't beep when imaging is finished.\n");
    printf("  --logfile=fn          -- Where to write a log. Specify /dev/null to suppress log.\n");
    printf("                           default is %s.\n",logfile);
    printf("  --info, -I            -- print configuration info.\n");

    bold("\nExisting File Options:\n");
    printf("  --append, -a          -- Append to existing file\n");
    printf("  --zap, -z             -- Erase outfile(s) before writing\n");

    bold("\nRaw Output Options:\n");
    printf("  --raw=fname, -r fname -- write block-by-block output to fname.\n");

    bold("\nAFF Options:\n");
    printf("  --outfile=fname, -ofname  -- write an AFF file.\n");
    printf("  --image_segsize=nnn, -S nnnn   -- set the AFF image_segsize (default %d)\n",
	   opt_datasegsize);
    printf("                     otherwise takes off where it left off.\n");
    printf("  --config=fn,  -c file -- Read question config from file (default %s)\n",opt_config);
    printf("  --make_config=fn      -- write a sample config file\n");
    printf("  --no_compress, -x     -- Do not compress. Useful on slow machines.\n");

    bold("\nError Recovery Options:\n");
    printf("  --error=0, -e0  -- Standard error recovery:\n");
    printf("                     Read disk 256K at a time until there are 5 errors in a row.\n");
    printf("                     Then go to the end of the disk and read backwards\n");
    printf("                     until there are 5 erros in a row. Then stop.\n");
    printf("  --error=1  -e1  -- Stop reading at first error.\n");
    printf("  --retry=nn -tnn -- change retry count from 5 to nn\n");
    printf("  --reverse, -v   -- Scan in reverse from the beginning.\n");

    bold("\nOther:\n");
    printf("  --help, -h      -- Print this message.\n");
    exit(0);
}


static struct option longopts[] = {
    { "raw",	       required_argument,  NULL, 'r'}, // can use this for flags, but not for longs
    { "outfile",       required_argument,  NULL, 'o'},
    { "quiet",         no_argument,        &opt_quiet, 'q'},
    { "silent",        no_argument,        NULL, 'Q'},
    { "readsectors",   required_argument,  NULL, 'R'},
    { "image_segsize", required_argument,  NULL, 'S'},
    { "zap",           no_argument,        &opt_zap, 'z'},
    { "config",        required_argument,  NULL, 'c'},
    { "help",          no_argument,        NULL, 'h'},
    { "retry",         required_argument,  NULL, 't'},
    { "no_compress",   no_argument,        NULL, 'x'},
    { "no_beeps",      no_argument,        NULL, 'B'},
    { "append",        no_argument,        NULL, 'a'},
    { "make_config",   required_argument,  NULL, 'm'},
    { "logfile",       required_argument,  NULL, 'l'},
    { "info",          no_argument,        NULL, 'I'},
    { "reverse",       no_argument,        NULL, 'v'},
    {0,0,0,0}
};


/****************************************************************
 *** isleep(): An informative sleep...
 ****************************************************************/
void isleep(int s)
{
  printf("isleep %d\n",s);
  for(int i=0;i<s;i++){
    printf("\rSleeping for %d seconds; %d left...",s,s-i);
    fflush(stdout);
    sleep(1);
  }
  printf("\r%50s\r","");
  fflush(stdout);
}


/****************************************************************
 *** Simson's termcap routines.
 ****************************************************************/

/* refresh():
 * Redraw the disk imaging status information...
 */
int rows = 0;
int cols = 0;
void nl() 
{
    tputs(clr_eol,0,putchar);
    tputs(carriage_return,0,putchar);
    tputs(newline,0,putchar);
}

void bold(const char *str)
{
    if(str[0]){
	tputs(enter_bold_mode,1,putchar);
	fputs(str,stdout);
	tputs(exit_attribute_mode,0,putchar);
    }
}


void gotorc(int row,int col)
{
    char buf[256];
    char *cc = buf;
    tputs(tgoto(tgetstr("cm",&cc),col,row),0,putchar);
}


void rightflush(int row,const char *str)
{
    gotorc(row,cols-(strlen(str)+1));
    fputs(str,stdout);
}




imager *im;
bool screen_cleared=false;
aff_ull total_bytes_compressed = 0;
aff_ull total_bytes_uncompressed = 0;

char *linebuf=0;

/* print val with commas */
void comma_printf(aff_ull val)
{
    char buf[64];
    char tmp[64];
    char t2[64];

    buf[0] = 0;
    if(val==0){
      strcpy(buf,"0");
    }

    while(val>0){
	int digits = val % 1000;		// get the residue
	val = val / 1000;		// and shift what's over

	if(val>0){			// we will still have more to do
	    sprintf(tmp,",%03d",digits);	// so pad it with zeros and a comma
	}
	else {
	    sprintf(tmp,"%d",digits);	// otherwise just get the value
	}
	strcpy(t2,buf);			// copy out the buffer
	strcpy(buf,tmp);		// copy in what we just did
	strcat(buf,t2);			// and put back what was there
    }
    fputs(buf,stdout);			// send it to standard output
}



void putchar_rev(char ch)
{
  tputs(enter_reverse_mode,0,putchar);
  putchar(ch);
  tputs(exit_attribute_mode,0,putchar);
}

void refresh()
{
    char timebuf[64];
    time_t now = time(0);
    strcpy(timebuf,ctime(&now));
    timebuf[25] = '\000';
    static int old_status_col = 0;

    if(screen_cleared == false){
	rows = tgetnum("li");
	cols = tgetnum("co");

	if(linebuf){
	    free(linebuf);
	    linebuf = (char *)malloc(cols+1);
	}

	tputs(clear_screen,0,putchar);
	screen_cleared = true;

	/* First time through, paint the stuff that doesn't change much */
    
	/* Source Device */
	int ncol = 16;		// where to put the first numeric column
	gotorc(1,1);   printf("Source device: ");
	gotorc(1,ncol); bold(infile);
	
	gotorc(2,   1);  printf("Model #:");
	gotorc(2,ncol);  bold(device_model);
	if(firmware_revision[0]){
	    gotorc(3,1);
	    printf("firmware: ");
	    gotorc(3,ncol);
	    bold(firmware_revision);
	}
	if(serial_number[0]){
	    gotorc(4,1);
	    printf("S/N:");
	    gotorc(4,ncol);
	    bold(serial_number);
	}
	
	/* Output */
	if(outfile_aff){
	    gotorc(1,40); printf("AFF Output: "); bold(outfile_aff);
	}
	if(outfile_raw){
	    gotorc(0,40); printf("Raw Output: "); bold(outfile_raw);
	}

	gotorc(3,40);  printf("Sector size:  %8d bytes",im->sector_size);nl();
	gotorc(4,40);  printf("Total sectors:");
	if(im->total_sectors){
	  comma_printf(im->total_sectors);
	}
	else {
	    printf("????");
	}
	nl();

	gotorc(10,0); putchar('[');
	for(int i=0;i<cols-2;i++){
	    putchar('-');
	}
	putchar(']');
    }

    /* Stuff that changes a lot */

    gotorc(0,30);
    if(im->imaging){
      tputs(enter_blink_mode,0,putchar);
      tputs(enter_reverse_mode,0,putchar);
      printf("IMAGING");
      tputs(exit_attribute_mode,0,putchar);
    }
    else {
      printf("       ");
    }

	
	
    /* Ornaments */
    rightflush(0,timebuf);
	
    gotorc(12,0);
    printf("Currently reading sector: ");
    comma_printf(im->last_sector_read);
    printf(" (%3d sectors at once) ",im->last_sectors_read);

    gotorc(14,0); 
    printf("SECTORS read: ");comma_printf(im->total_sectors_read);
    printf(" (%5.2f%%) ",
	   im->total_sectors_read*100.0/(float)im->total_sectors);

    printf(" blank: ");comma_printf(im->total_blank_sectors);
    if(im->total_sectors_read == im->total_blank_sectors){
      printf("  NO DATA YET!");
    }
    nl();

    gotorc(15,0);
    printf("BAD SECTORS: ");comma_printf(im->bad_sectors_read);
    if(im->errors_in_a_row){
	printf(" (%d errors in a row)",im->errors_in_a_row);
    }
    nl();

    gotorc(17,0);


#ifdef __FreeBSD__    
    int outfd = 0;
    if(im->af) outfd = fileno(im->af->aseg);
    if(im->fout) outfd = fileno(im->fout);
    struct statfs buf;
    if(fstatfs(outfd,&buf)==0){
      gotorc(16,0);   printf("Free space: ");
      aff_ull free_kbytes = (unsigned long long)buf.f_bfree / 2;
      aff_ull free_mbytes = free_kbytes / 1024;
      comma_printf(free_mbytes);
      printf(" MB (%5.2f%%)", buf.f_bfree*100.0/buf.f_blocks); 
    }
#endif    

    if(opt_nocompress==0){
	gotorc(18,0);
	printf("         Total bytes read: ");
	comma_printf(total_bytes_uncompressed);
	gotorc(19,0);
	printf(" Compressed bytes written: ");
	comma_printf(total_bytes_compressed);
	if(compressing){bold("COMPRESSING... ");}
	nl();
	gotorc(20,0);
	if(total_bytes_uncompressed>0){
	  printf("Overall compression ratio: %2.6f%%",
		 (100.0 * (total_bytes_uncompressed-total_bytes_compressed))/
		 total_bytes_uncompressed);
	}
	nl();
    }


    //printf("Time spent reading:     %s");nl();
    //printf("Time spent compressing: %s");nl();

    
    if(im->total_sectors){
	int new_status_col = (int)(((float)im->last_sector_read
				    /im->total_sectors) * (cols-2)) + 1;
	if(old_status_col && old_status_col != new_status_col){
	    /* Need to erase old status */
	    gotorc(10,old_status_col);
	    putchar_rev('=');
	}
	old_status_col = new_status_col;

	gotorc(10,new_status_col);
	if(im->last_direction==1){
	  putchar_rev('>');
	}
	else{
	  putchar_rev('<');
	}
    }
    gotorc(20,0);		// leave the cursor here for errors
    fflush(stdout);
    }


void compression_callback(int bytes_uncompressed,int bytes_compressed)
{
  compressing = (bytes_uncompressed==0 && bytes_compressed==0);
    total_bytes_compressed  += bytes_compressed;
    total_bytes_uncompressed  += bytes_uncompressed;
    if(opt_quiet==0 && opt_silent==0 && finishing==false){
      refresh();
    }
}



/* sig_intr:
 * Close the AFF file if possible
 * If this is the second time we were called, just exit...
 */

int depth = 0;
void sig_intr(int arg)
{
    if(++depth==1){
      // don't get confused...
	af_enable_compression(im->af, !opt_nocompress, 0);
	gotorc(rows-4,0);
	printf("Interrupt!\n");
	printf("\n\n\n");
	printf("Closing output file...\n");
	if(im->af){
	    if(af_close(im->af)){
		warn("Can't close file '%s'\n",af_filename(im->af));
	    }
	}
    }
    fflush(stdout);
    exit(1);
}


char lastchar(const char *str)
{
    return str[strlen(str)-1];
}


/* Special logic for opening a FreeBSD device... */
void Sleep(int msec)
{
    usleep(msec*1000);
}



void make_config(char *fname)
{
    if(access(fname,F_OK)==0){
	fprintf(stderr,"make_config: %s exists.\n",fname);
	exit(1);
    }
    FILE *f = fopen(fname,"w");
    if(!f) err(1,"make_config(%s) ",fname);
    fputs("#\n",f);
    fputs("#\n",f);
    fputs("#\n",f);
    fputs("# Sample config file for aimage...\n",f);
    fputs("# Syntax:  ask <segname> <question>\n",f);
    fputs("#\n",f);
    fputs("ask name Your Name:\n",f);
    fputs("ask date_acquired Date this drive was acquired (yyyy-mm-dd):\n",f);
    fputs("ask case_number Case Number:\n",f);
    fputs("#\n",f);
    fputs("# Feel free to add your own!\n",f);
    fclose(f);
    printf("%s created\n",fname);
    exit(0);
}


void process_config(AFFILE *af)
{
    FILE *f = fopen(opt_config,"r");	// open the config file to find questions to ask
    if(!f) return;			// get back to the user
    while(!feof(f)){
	char buf[1024];			// a righteous buffer for questions and answers
	char *sep = " \t";
	char *last;
	memset(buf,0,sizeof(buf));
	fgets(buf,sizeof(buf)-1,f);
	
	char *cc = index(buf,'\n');
	if(cc) *cc = '\000';		// remove the \n

	char *cmd = strtok_r(buf,sep,&last);

	if(!cmd || strcmp(cmd,"ask")!=0) continue; // not an ask command
	char *segname = strtok_r(NULL,sep,&last);
	if(!segname){
	    fprintf(stderr,"error in config file. No segnament name in: %s\n",buf);
	    continue;
	}
	
	/* Ask the question and get the response */
	char *val = readline(last);

	/* And write the response into the segment */
	af_put_seg(af,segname,0,val,strlen(val));
	free(val);
    }
    fclose(f);
}



/*
 * Try to open the device *device, and then
 * change *device to be the name of the actual device
 * that was opened...
 *
 * Allow the user to specify ATA channel 0 as either
 * ata0 or ide0.
 */

int  open_dev(char **device)
{
    int atadev = -1;
    int fd;

    sscanf(*device,"ata%d",&atadev);	// try to find ata0
    if(atadev==-1){
	sscanf(*device,"ide%d",&atadev);	// try to find ide0
    }

    if(atadev != -1){			// if we found the device
	/* Create the attach and detach commands */

	

	char dev0[64];			// space for the first channel
	char dev1[64];			// space for the second channel
	char *dev[2] = {dev0,dev1};

	make_attach_commands(cmd_attach,cmd_detach,dev0,dev1,atadev);
	system(cmd_detach);	// make sure we are detached first

	int i;
	for(i=0;i<10;i++){
	    int delay = i*3;
	    printf("\nOpening special ATA Bus #%d...\n",atadev);
	    if(i>0){
		printf("Attempt %d out of %d.\n",i+1,10);
	    }
	    printf("# %s\n",cmd_attach);
	    system(cmd_attach);
	    
	    /* See if we found the device */
	    for(int d=0;d<2;d++){
		if(access(dev[d],F_OK)==0){
		    if(access(dev[d],R_OK)){
			// don't have permission to read it.
			// this is bad
			err(1,dev[d]);
		    }
		    if(delay){
			printf("Waiting %d second%s for %s to spin up...\n",
			       delay,delay==1?"":"s",dev[d]);  
			isleep(delay);
		    }
		    fd = open(dev[d],O_RDONLY);
		    if(fd>0){
			/* The device was successfully opened.
			 * Remember dev[d] and return it
			 */
			*device = strdup(dev[d]);
			return fd;		// got it!
		    }
		    perror(dev[d]);		// give a warning
		}
	    }
	    printf("Detaching device and trying again...\n");
	    printf("# %s\n",cmd_detach);
	    system(cmd_detach);
	    isleep(delay);
	}
    }
    return -1;
}


void print_info()
{
    for(int i=0;i<2;i++){
	char master[64];
	char slave[64];
	printf("Device %d:\n",i);
	make_attach_commands(cmd_attach,cmd_detach,master,slave,i);
	printf("  attach with '%s'\n",cmd_attach);
	printf("  deatch with '%s'\n",cmd_detach);
	printf("  master is %s\n",master);
	printf("  slave is %s\n",slave);
	printf("\n");
    }
    exit(0);
}

int main(int argc,char **argv)
{
    setupterm((char *)0,1,(int *)0);
    progname = argv[0];

    /* Create a logfile... */
    char buf[1024];
    sprintf(buf,"%s/.aimage_log",getenv("HOME"));
    logfile = strdup(buf);

    /* build the option string */
    char optstring[256];
    optstring[0] = 0;
    int i;
    char *cc = optstring;
    for(i=0;longopts[i].name;i++){
	switch(longopts[i].has_arg){
	case no_argument:
	    *cc++ = longopts[i].val;
	    break;
	case required_argument:
	case optional_argument:
	    *cc++ = longopts[i].val;
	    *cc++ =':';
	    break;
	default:
	    printf("unknown value for longopts[%d].has_arg=%d\n",
		   i,longopts[i].has_arg);
	    exit(9);
	}
    }
    *cc++ = 0;

    int ch;
    while ((ch = getopt_long_only(argc, argv,
				  optstring,longopts,NULL)) != -1) {
	switch (ch) {
	case 'r': outfile_raw = optarg;	    break;
	case 'o': outfile_aff = optarg;	    break;

	case 'q': opt_quiet++;		break; 
	case 'Q': opt_silent++;		break;
	case 'R':
	    opt_readsectors = atoi(optarg);
	    break;
	case 'z': opt_zap ++;		    break;
	case 'a': opt_append ++;	    break;
	case 'k':
	    opt_skip  = atoi(optarg);
	    if(lastchar(optarg)=='s'){
		opt_skip_sectors = 1;
	    }
	    break;
	case 'S':
	    opt_datasegsize = atoi(optarg);
	    break;
	case 'c':
	    opt_config = optarg;
	    break;
	case 'x':
	    opt_nocompress = 1;
	    break;
	case 'e':
	    opt_error = atoi(optarg);
	    break;
	case 't':
	  opt_retry = atoi(optarg);
	  break;
	case 'B':
	  opt_beeps = 0;
	  break;

	case 'm':
	    make_config(optarg);
	    break;
	    
	case 'I':
	    print_info();
	    break;

	case 'v':
	    opt_reverse =1;
	    break;

	case 'h':
	case '?':
	default:
	    usage();
	case 0:
	    break;
	}
    }

    argc -= optind;
    argv += optind;

    if(argc<1){
	usage();
    }
    infile = argv[0];

    argc--;
    argv++;

    /* IF there is an additional output file,
     * and outfile_aff has not been specified, use that
     */
    if(outfile_aff==0 && argc>0){
	outfile_aff = argv[0];
	argc--;
	argv++;
    }

    start_time = time(0);		// excellent!

    if(opt_zap){
	if(outfile_raw) unlink(outfile_raw);	// if we are zapping
	if(outfile_aff) unlink(outfile_aff);
    }

    /* If either file exists and we are not appending, give an error */
    if(!opt_append){
	if(access(outfile_raw,F_OK)==0) errx(1,"%s: file exists",outfile_raw);
	if(access(outfile_aff,F_OK)==0) errx(1,"%s: file exists",outfile_aff);
    }

    /* First, open the input file.
     * If the file specified by the user works, just open it.
     * If the file doesn't work, see if it is a operating system
     * specific filename such as "ide0" or "ata0", in which the
     * operating system-specific code will attempt to attach the
     * device.
     */

    int ifd = open(infile,O_RDONLY);
    if(ifd<0){				
	/* Attempt to open infile failed; check for a special
	 * device name...
	 */
	ifd = open_dev(&infile);
    }

    /* If we haven't been able to open the infile, give up. */
    if(ifd<0){
	perror(infile);
	exit(1);
    }






    /* If the user is imaging to an AFF file,
     * open it and try to ident the drive.
     * Drive ident is not done if writing to a raw file, because
     * there is no place to store the ident information. This will be changed
     * when we can write an XML log.
     */
    AFFILE *af=0;
    if(outfile_aff){
	if(af_identify(outfile_aff)==0){
	    fprintf(stderr,"%s exists and is not an AFF file.\n",outfile_aff);
	    fprintf(stderr,"Delete it or move it first.\n");
	    exit(-1);
	}

	af = af_open(outfile_aff,O_CREAT|O_RDWR,0666);
	if(!af){
	    err(1,"af_open %s: ",outfile_aff);
	}
	/* If the segment size hasn't been set, then set it */
	af_set_sizes(af,opt_datasegsize,(unsigned long long)0);		// 
	af_enable_writing(af,1);	// we will be writing this file
	af_enable_compression(af,
			      !opt_nocompress, compression_callback);
	process_config(af);		// process the config file for questions
	ident(infile,af);		// ident the drive
	/* TK: date_acquired, accession_gid */
    }

    im = new imager();
    im->retry_count = opt_retry;

    if(opt_quiet ==0 && opt_silent==0){
	im->status = refresh;
    }

    if(im->set_input(ifd)){
      af_close(af);	// whoops
      exit(0);
    }

    if(af) im->set_af(af);

    /* See if the skipping makes sense */
    if(!opt_skip_sectors){
	if(opt_skip % im->sector_size != 0){
	    fprintf(stderr,
		    "Skipping must be an integral multiple of sector size "
		    "(%d bytes)\n",im->sector_size);
	    af_close(af);
	    exit(1);
	}
	opt_skip /= im->sector_size;	// get the actuall offset
    }


    FILE *fout = 0;
    if(outfile_raw){
	fout = fopen(outfile_raw,"wb");
	if(!fout){
	    err(1,"fopen(%s)",outfile_raw);
	}
	im->set_fout(fout);
    }

    int starting_direction = 1;
    if(opt_reverse) starting_direction = -1;

    if(opt_beeps) putchar('\007');			// beep
    signal(SIGINT,sig_intr);	// set the signal handler
    im->hash_setup();		// get ready...
    im->image_loop(opt_skip,im->total_sectors,starting_direction,
		   opt_readsectors,opt_error); // start the process
    signal(SIGINT,0);		// unset the handler

    /* Finish up */
    finishing = true;
    gotorc(rows-2,0);
    tputs(clr_eos,0,putchar);
    printf("\n\n");

    char print_buf[256];
    unsigned char md5_buf[16];
    MD5_Final(md5_buf,&im->md5);
    printf("raw image md5:  %s\n",
	   af_hexbuf(print_buf,sizeof(print_buf),md5_buf,16,1));

    unsigned char sha1_buf[20];
    SHA1_Final(sha1_buf,&im->sha);
    printf("raw image sha1: %s\n",
	   af_hexbuf(print_buf,sizeof(print_buf),sha1_buf,20,1));

    if(fout){
	if(fclose(fout)){
	    warn("fclose(%s)",outfile_raw);
	    fprintf(stderr,
		    "This shouldn't happen. '%s' may be corrupt.\n",
		    outfile_raw);
	}
	fout=0;
    }

    /* AFF Cleanup... */
    if(af){
	if(af_update_seg(af,AF_MD5,0,md5_buf,16,0)){
	    perror("Could not update AF_MD5");
	}
	if(af_update_seg(af,AF_SHA1,0,sha1_buf,20,0)){
	    perror("Could not update AF_SHA1");
	}
	if(af_update_segq(af,AF_BADSECTORS, im->bad_sectors_read,1)){
	    perror("Could not update AF_BADSECTORS");
	}
	if(af_update_segq(af,AF_BLANKSECTORS, im->total_blank_sectors,1)){
	    perror("Could not update AF_BLANKSECTORS");
	}

	if(af_close(af)){
	    perror("af_close");
	    fprintf(stderr,"af_close failed. This shouldn't happen.\n");
	    fprintf(stderr,"Run 'ainfo -v %s' to see if %s is corrupt.\n",
		    outfile_aff,outfile_aff);
	}
	af = 0;

	/* Let's make the user feel good... */
	if(im->total_bytes_read>0){
	    printf("Compression ratio: %4.1f%%\n",im->compression_ratio());
	}
    }

#ifdef __FreeBSD__
    /* Need to detach if we are using freebsd and attached */
    if(cmd_detach[0]){
	printf("Detaching drive...\n");
	system(cmd_detach);
    }
#endif    
    if(opt_beeps && !opt_quiet && !opt_silent){
      for(int beep=0;beep<4;beep++){
	putchar('\007');
	fflush(stdout);
	Sleep(300);
      }
    }

    return 0;
}
